# = Image Operation Cancelling Tool
#   Copyright FUJITSU LIMITED 2011
#
require 'pathname'
require 'rubygems'
require 'socket'
require 'rexml/document'
require 'thread'

OS_TYPE = ENV['IMAGECANCELLER_RCX_OS_TYPE']
RUBY_TYPE = ENV['IMAGECANCELLER_RCX_RUBY_TYPE']

# = ImageCancellerParams -- defined parameters for this tool
module ImageCancellerParams
  SCWSH_WIN = "/ScwPro/bin/scwsh.exe"
  SCWSH_LIN = "/FJSVscw-deploysv/sys/scwsh"
  MGRDIR_WIN = "/Manager"
  MGRDIR_LIN = "/FJSVrcvmr"
  INSTDIR_WIN = "IMAGECANCELER_RCX_PATH"
  INSTDIR_LIN = "/opt"
  LOG_DIR_WIN = "/Manager/var/log"
  LOG_DIR_LIN = "/../../var/opt/FJSVrcvmr/log"
  LOG_FILE_NAME = "/rcximagecanceller.log"
  LOG_FILE_BACK_NAME = "#{LOG_FILE_NAME}.bak"
  LOG_LEVEL_INFO = "INFO"
  LOG_LEVEL_ERROR = "ERROR"
  LOG_SIZE_MAX = 1024 * 1024
  LOCK_FILE = "/imagecanceller"
  LOCK_EXT = ".lck"
  SCWSH_CON_PREFIX = "RCX"
  STATUS_CANCELLED = /^Cancelled$/
  STATUS_RUNNING = /^Running$/
  CMD_BACKUP = /^clone-backup.1$/
  COMPUTER_GROUP = "PG:\\rcx\\"
  WAIT_TIME = 60
  WAIT_INTERVAL = 2
  CMD_ERROR  = -1
  MSWIN = /mswin/
  LINUX = /linux/
  JRUBY = /jruby/
  RCOMMAND_CLASS  = 'com.fujitsu.rcx.vio.awwn.rmc.ProcessExecRMCFacade'
  RCOMMAND_METHOD = 'execute'
end

# ImageCancellerMessage -- defined rcx message for image canceller
module ImageCancellerMessage
  ERR67101 = "FJSVrcx:ERROR:67101:not privileged"
  ERR67102 = "FJSVrcx:ERROR:67102:#{ARGV[0]}:illegal option\nUsage:rcximagecancel -server resource"
  ERR67162 = "FJSVrcx:ERROR:67162:#{ARGV[0]}:is required\nUsage:rcximagecancel -server resource"
  ERR67130 = "FJSVrcx:ERROR:67130:#{ARGV[0]}:option requires an argument\nUsage:rcximagecancel -server resource"
  ERR67137 = "FJSVrcx:ERROR:67137:command is already running"
  ERR67154 = "FJSVrcx:ERROR:67154:#{ARGV[1]}:not found"
  ERR68253 = "FJSVrcx:ERROR:68253:A timeout occurred during cancel the server(#{ARGV[1]}) in the image operation."
  ERR00001 = "FJSVrcx:ERROR:00001:#{ARGV[1]}:image operation isn't running"
  ERR67999 = "FJSVrcx:ERROR:67999:internal error, failed to execute cancel."
  ERR00002 = "FJSVrcx:ERROR:00002:command execution error."
end

# ImageCancellerError -- exception for image canceller
class ImageCancellerError < Exception
  attr_accessor :message
  def initialize(num)
    @message=num
  end
end
class ImageCancellerAuthorityError < ImageCancellerError
end
class ImageCancellerFormatError < ImageCancellerError
end
class ImageCancellerNoServerError < ImageCancellerError
end
class ImageCancellerRequestError < ImageCancellerError
end
class ImageCancellerLockError < ImageCancellerError
end
class ImageCancellerInternalError < ImageCancellerError
end
class ImageCancellerTimeoutError < ImageCancellerError
end
class ImageCancellerCommandExecutionError < ImageCancellerError
end

# ImageCancellerDir -- get directory of manager and command
class ImageCancellerDir
  include ImageCancellerParams
  
  if OS_TYPE =~ MSWIN
    MANAGER_DIR  = MGRDIR_WIN
    LOG_DIR = LOG_DIR_WIN
  else
    MANAGER_DIR  = MGRDIR_LIN
    LOG_DIR = LOG_DIR_LIN
  end
  
  def self.install_dir
    if OS_TYPE =~ MSWIN
        return ENV[INSTDIR_WIN]
    else
      return INSTDIR_LIN
    end
  end
  
  def self.manager_dir
    return Pathname.new(install_dir + MANAGER_DIR).realpath.to_s
  end
  
  def self.log_dir
    return Pathname.new(install_dir + LOG_DIR).realpath.to_s
  end
  
  def self.scwsh
    if OS_TYPE =~ MSWIN
      return "#{install_dir}#{SCWSH_WIN}"
    else
      return "#{install_dir}#{SCWSH_LIN}"
    end
  end
end

# === Copy from RCVE V2.1.3 Manager - start
# ImageCancellerSJRMCParams -- defined parameters for SJRMC
module ImageCancellerSJRMCParams
  # default port to connect
  DEFAULT_PORT = 23460  
  # Key to look up services file.
  SERVICES_KEY = "rcxmgr"  
end

# ImageCancellerRcxRMCAuthentication -- rmc authentication
class ImageCancellerRcxRMCAuthentication

  RMC_SECRET_KEY_FILE = File.expand_path("#{ImageCancellerDir.manager_dir}/rails/db/rmc_key")
  DEFAULT_KEY = '0'
  
  # === Get the rmc session id
  # ==== Parameters & return value
  # *Return* :: [String] Session id
  def self.secret_key
      @@skey ||= get_secret_key
      return @@skey
  end

  # === Authentication
  # ==== Parameters & return value
  # _secret_key_ :: [String] secret_key
  # ==== Exceptions
  # SecurityError :: authentication failed
  def self.authenticate(secret_key)
    @@skey ||= DEFAULT_KEY
    if secret_key != @@skey
      raise SecurityError, 'authehtication failed', caller
    end
  end

  private   

  def self.update_secret_key
    @@skey = get_secret_key
  end
  
  def self.get_secret_key
    key=nil
    begin
      File.open(RMC_SECRET_KEY_FILE, "r") do |fd|
        key = fd.read
      end
    rescue Exception => e
      # Can't log here because RcxLogger is depending on ImageCancellerRcxRMCAuthentication
      # below code is for coverage
      key = DEFAULT_KEY
    end
    key ||= DEFAULT_KEY
    return key
  end
end

# ImageCancellerRcxUtil -- 
class ImageCancellerRcxUtil  
  def self.get_service_port(name, type, default_port)
    return Socket.getservbyname(name, type) ||  default_port
  rescue
    return default_port
  end
end

# = ImageCancellerSJRMC -- Simple Java Remote Method Call
module ImageCancellerSJRMC
  include REXML
  
  def call(class_name, method_name, args)
    begin
      Thread.exclusive do
        ImageCancellerSJRMC::SJRMCSocket.socket = TCPSocket.open('127.0.0.1', get_port) unless ImageCancellerSJRMC::SJRMCSocket.socket
        cmd = createXMLCmd(class_name, method_name, args)
        ImageCancellerSJRMC::SJRMCSocket.socket.puts(cmd)
        result = ImageCancellerSJRMC::SJRMCSocket.socket.gets
        return parse_result(result)
      end
    rescue Exception
      begin
        ImageCancellerSJRMC::SJRMCSocket.socket.close if ImageCancellerSJRMC::SJRMCSocket.socket
      rescue Exception
      end
      ImageCancellerSJRMC::SJRMCSocket.socket = nil
      raise
    end
  end
  
  private

  class SJRMCSocket
    def self.socket=(socket)
      @@socket = socket
    end
    
    def self.socket
      @@socket ||= nil
    end
  end

  class SJRMCPort
    def self.no
      @@no ||= ImageCancellerRcxUtil.get_service_port(ImageCancellerSJRMCParams::SERVICES_KEY, "tcp", ImageCancellerSJRMCParams::DEFAULT_PORT)
      return @@no
    end
  end
  
  def get_port()
    return SJRMCPort.no
  end

  def createXMLCmd(class_name, method_name, args)
    doc = REXML::Document.new
    cmd = REXML::Element.new('Command',doc)
    cmd.add_attribute('skey', ImageCancellerRcxRMCAuthentication.secret_key)
    cmd.add_attribute('class', class_name)
    cmd.add_attribute('method', method_name)
    
    if args
      args.each do |arg|
        _arg = REXML::Element.new('Arg', cmd)
        arg.to_s.split(/\x0d\x0a|[\x0d\x0a]/).each do |line|
          l =  REXML::Element.new('P', _arg)
          l.add_text(line.gsub(/[\x00-\x1f]/, ' '))
        end
      end
    end
    
    output=''
    doc.write(output, -1)
    return output;
  end
  
  def parse_result(resultStr)
    doc = REXML::Document.new resultStr
    messages = []
    elements = doc.get_elements("/CommandResult/Message")
    elements.each do | permsg |
      msgs =  permsg.get_elements("P").collect do |line|
        line.text
      end
      messages << msgs.join("\n")
    end
    
    if (elements.length == 0) 
      raise ImageCancellerSJRMCException.new("SJRMC protocol exception. /CommandResult/Message not found")
    end
    errorCode = doc.get_text("/CommandResult/ErrorCode").to_s
    if (errorCode == '')
      raise ImageCancellerSJRMCException.new("SJRMC protocol exception. /CommandResult/ErrorCode not found")
    end
    return SJRMCResult.new(errorCode.to_i, messages)
  end
  
  # == Result of SJRMC.call(include errorcode and message)
  class SJRMCResult
    attr_reader :error_code, :messages
    
    # ==== Parameters & return value
    # _error_code_:: [Integer] Error code
    # _messages_::    [String] Messages
    def initialize(error_code, messages)
      @error_code = error_code.to_i
      @messages = messages
    end

    # === This method exists to be compatible with old version
    # ==== Parameters & return value
    # *Return*:: [String] Message at index 0
    def message
      return @messages[0]  
    end
    
    # ==== Parameters & return value
    # *Return* :: [String] String format #{Error code}:#{Error message,...}
    def to_str
      return "#{@error_code}:#{@messages.join(', ')}"
    end
  end
end

class ImageCancellerSJRMCException < Exception
end
# === Copy from RCVE V2.1.3 Manager - end

# ImageCancellerLogger --- logger for image canceller
class ImageCancellerLogger
  include ImageCancellerParams
  LOG_FILE = "#{ImageCancellerDir.log_dir}#{LOG_FILE_NAME}"
  LOG_FILE_BACK = "#{ImageCancellerDir.log_dir}#{LOG_FILE_BACK_NAME}"

  def self.info(message)
    write_log(message, LOG_LEVEL_INFO)
  end

  def self.error(message)
    write_log(message, LOG_LEVEL_ERROR)
  end

  def self.write_log(message, level)
    mode = "a"
    begin
      if File.exists?(LOG_FILE)
        s = File::stat(LOG_FILE)
        if s.size > LOG_SIZE_MAX
          File.rename(LOG_FILE, LOG_FILE_BACK)
          mode = "w"
        end
      else
        mode = "w"
      end
      File.open(LOG_FILE, mode) do |f|
        f.puts "[#{Time.now}]#{level}:#{message}"
      end
    rescue Exception => ex
      # File Access Error
    end
  end
end

# = ImageCancellerScwCommand -- utilities for scwsh command exec
if RUBY_TYPE =~ ImageCancellerParams::JRUBY
  require 'java'
  require "#{ImageCancellerDir.manager_dir}/opt/FJSVrcxmr/lib/plugin/awwn.jar"
end

class ImageCancellerScwCommand 
  include ImageCancellerParams
  include ImageCancellerSJRMC
  include ImageCancellerMessage
  def initialize(server_name)
    @connection = SCWSH_CON_PREFIX + server_name
  end
  
  def exec(command, success_code=0)
    cmd = [ImageCancellerDir::scwsh] + ["-b:#{@connection}", "#{command}"]
    begin
      if RUBY_TYPE =~ JRUBY
        rmc_facade = com.fujitsu.rcx.vio.awwn.rmc.ProcessExecRMCFacade.new
        command_result = rmc_facade.execute(cmd.to_java(:string))
        return_code    = command_result.getErrorCode
        output         = command_result.getMessags[0]
        stderr_message = command_result.getMessags[1]
      else
        ret = call(RCOMMAND_CLASS, RCOMMAND_METHOD, cmd)
        output = ret.message
        return_code = ret.error_code
        stderr_message = ret.messages[1]
      end
    rescue Exception => ex
      ImageCancellerLogger.error("ImageCancellerScwCommand.exec(): failed to exec cmd. cmd=#{command}, error=#{ex.class}/#{ex.message}")
      return_code = CMD_ERROR
      stderr_message = ex.message
    end
    if return_code != success_code
      ImageCancellerLogger.error("ImageCancellerScwCommand.exec(): failed to exec cmd. cmd=#{command}, return_code=#{return_code}, stderr_message=#{stderr_message}")
      raise ImageCancellerCommandExecutionError.new(ERR00002)
    end
    return output
  end
  
  def get_vals(command)
    output = exec(command)
    return [] if output.nil? || output.empty?
    values = output.split
    return [] if values.size >= 2 && ( values[1].nil? || values[1].empty? )
    return values
  end
end

# ImageCancellerLock -- lock system for scwsh request
class ImageCancellerLock
  include ImageCancellerParams
  include ImageCancellerMessage
  def initialize(target)
    @target = target
  end
  
  def lock(&block)
    fname = "#{File.dirname(__FILE__)}#{LOCK_FILE}#{@target}#{LOCK_EXT}"
    begin
      File.open(fname, "w") do |f|
        begin
          lock_state = f.flock(File::LOCK_EX | File::LOCK_NB)
          raise ImageCancellerLockError.new(ERR67137) unless lock_state
          block.call
        ensure
          f.flock(File::LOCK_UN)
        end
      end
      File.delete(fname)
    rescue ImageCancellerLockError => ex
      raise ex
    rescue Errno::EACCES => ex
      # failed to delete the lock file because the file is opened
      raise ex
    rescue Exception => ex
      File.delete(fname) if File.exist?(fname)
      ImageCancellerLogger.error("ImageCancellerLock.lock(): failed to lock. error=#{ex.class}/#{ex.message}")
      raise ex
    end
  end
end

# ImageCancellerMain -- main procedure of image canceller
if OS_TYPE =~ ImageCancellerParams::MSWIN
  require 'Win32API' unless RUBY_TYPE =~ ImageCancellerParams::JRUBY
end
class ImageCancellerMain
  include ImageCancellerParams
  include ImageCancellerMessage
  def initialize(args)
    @scwsh=nil
    @request_id = nil
    @args = args
    @input_server_name = nil
  end

  def admin_check
    begin 
      raise ImageCancellerAuthorityError.new(ERR67101) if !is_admin?
    rescue ImageCancellerAuthorityError => ex
      raise ex
    rescue Exception => ex
      raise ImageCancellerInternalError.new(ERR67999)
    end
  end

  def params_check
    begin 
      raise ImageCancellerFormatError.new(ERR67162) if @args.length == 0
      raise ImageCancellerFormatError.new(ERR67102) if @args[0] != "-server"
      raise ImageCancellerFormatError.new(ERR67130) if @args[1] == nil
      raise ImageCancellerFormatError.new(ERR67102) if @args.length > 2
      @input_server_name = @args[1]
    rescue ImageCancellerFormatError => ex
      ImageCancellerLogger.error("ImageCancellerMain.params_check(): parameter error. args=#{@args}, error=#{ex.class}/#{ex.message}")
      raise ex
    rescue Exception => ex
      ImageCancellerLogger.error("ImageCancellerMain.params_check(): parameter error. args=#{@args}, error=#{ex.class}/#{ex.message}")
      raise ImageCancellerInternalError.new(ERR67999)
    end
  end

  def execute
    begin
      @scwsh=ImageCancellerScwCommand.new(@input_server_name)
      server_lock do
        logon do
          check_server_exists
          get_request_id
          request_lock do
            select_request_id
            pre_check_cancel
            execute_cancel
            wait_cancel
          end
        end
      end
    rescue ImageCancellerError => ex
      ImageCancellerLogger.error("ImageCancellerMain.execute(): failed to execute cancel. error=#{ex.class}/#{ex.message}")
      raise ex
    rescue Exception => ex
      ImageCancellerLogger.error("ImageCancellerMain.execute(): failed to execute cancel. error=#{ex.class}/#{ex.message}")
      raise ImageCancellerInternalError.new(ERR67999)
    end
	end

private

  def is_admin?
    is_user_an_admin = ENV["RCX_IS_ADMIN"]
    if OS_TYPE =~ MSWIN
	  unless RUBY_TYPE =~ ImageCancellerParams::JRUBY
        is_user_an_admin_ruby = Win32API.new('shell32', 'IsUserAnAdmin', '', 'i')
        return false if is_user_an_admin_ruby.call() == 0
        return true
      end
    end
    is_user_an_admin = is_user_an_admin.chomp.to_i
    return false if is_user_an_admin != 0
    return true
  end

  def logon(&block)
    begin
      ImageCancellerLogger.info("logon")
      @scwsh.exec("logon")
      @scwsh.exec("computer")
      block.call
    rescue Exception => ex
      ImageCancellerLogger.error("ImageCancellerMain.logon(): failed to logon. error=#{ex.class}/#{ex.message}")
      raise ex
    ensure
      ImageCancellerLogger.info("logoff")
      @scwsh.exec("logoff")
      @scwsh.exec("quit")
    end
  end
  
  def check_server_exists
    server_name = nil
    begin
      ImageCancellerLogger.info("check server exists -> #{@input_server_name}")
      server_name = @scwsh.exec("select name=#{COMPUTER_GROUP}#{@input_server_name}")
    rescue Exception => ex
      ImageCancellerLogger.error("ImageCancellerMain.check_server_exists(): server not found. server=#{@input_server_name}, error=#{ex.class}/#{ex.message}")
      raise ImageCancellerNoServerError.new(ERR67154) unless server_name
      raise ex
    end
  end
  
  def get_request_id
    ImageCancellerLogger.info("get request id")
    @request_id = nil
    @request_id = @scwsh.get_vals("param list .sysinfo.lastbatch.rid")[1]
    raise ImageCancellerRequestError.new(ERR00001) unless @request_id
  end

  def request_lock(&block)
    ImageCancellerLogger.info("request lock")
    ImageCancellerLock.new(@request_id).lock(&block)
  end

  def server_lock(&block)
    ImageCancellerLogger.info("server lock")
    ImageCancellerLock.new(@input_server_name).lock(&block)
  end
  
  def select_request_id
    begin
      ImageCancellerLogger.info("select request id -> #{@request_id}")
      @scwsh.exec("request")
      @scwsh.exec("select #{@request_id}")
    rescue Exception => ex
      ImageCancellerLogger.error("ImageCancellerMain.select_request_id(): failed to select request. request_id=#{@request_id}, error=#{ex.class}/#{ex.message}")
      raise ImageCancellerRequestError.new(ERR00001)
    end
  end
  
  def pre_check_cancel
    ImageCancellerLogger.info("pre check cancel")
    params = @scwsh.get_vals("result list columns=@status")
    status = params[0]
    if status &&  status =~ STATUS_RUNNING
      return
    else
      ImageCancellerLogger.error("ImageCancellerMain.pre_check_cancel(): cancel pre-check error. status=#{status}")
      raise ImageCancellerRequestError.new(ERR00001)
    end
  end
  
  def execute_cancel
    ImageCancellerLogger.info("execute cancel")
    @scwsh.exec("request")
    @scwsh.exec("select #{@request_id}")
    @scwsh.exec("cancel")
  end
  
  def wait_cancel
    ImageCancellerLogger.info("wait cancel")
    current_time = Time.now
	time, status = nil, nil
    begin
      loop do
        time = Time.now - current_time
        if time > WAIT_TIME
          raise ImageCancellerTimeoutError.new(ERR68253)
        end
        status = @scwsh.get_vals("result list columns=@status")[1]
        ImageCancellerLogger.info("time = #{time}, status = #{status}")
        return if status =~ STATUS_CANCELLED
        sleep WAIT_INTERVAL
      end
    rescue ImageCancellerTimeoutError => ex
      ImageCancellerLogger.error("ImageCancellerMain.wait_cancel(): timeout occurred. status=#{status}, time=#{time}, error=#{ex.class}/#{ex.message}")
      raise ex
    end
  end  
end

begin
  canceller = ImageCancellerMain.new(ARGV)
  canceller.admin_check
  canceller.params_check
  canceller.execute
rescue Exception => ex
  if ex.class != ImageCancellerAuthorityError
    ImageCancellerLogger.error("Execute cancel tool failed. error=#{ex.class}/#{ex.message}")
    ImageCancellerLogger.error("detail = \n#{ex.backtrace.join("\n")}") if ex.backtrace
  end
  puts ex.message
  exit(1)
end